import datetime
import json
from unittest.mock import MagicMock
from unittest.mock import patch

from django.core.urlresolvers import reverse
from django.utils import timezone

from orchestra.core.errors import TimerError
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import TaskTimer
from orchestra.models import TimeEntry
from orchestra.models import Worker
from orchestra.project_api.serializers import TaskTimerSerializer
from orchestra.project_api.serializers import TimeEntrySerializer
from orchestra.tests.helpers import EndpointTestCase
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.load_json import load_encoded_json
from orchestra.views import bad_request
from orchestra.views import forbidden
from orchestra.views import internal_server_error
from orchestra.views import not_found


class TimeEntriesEndpointTests(EndpointTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.url = reverse('orchestra:orchestra:time_entries')
        self.tasks = Task.objects.filter(assignments__worker=self.worker)
        self.task = self.tasks[0]
        self.assignment = (self.task.assignments.filter(worker=self.worker)
                           .first())
        self.time_entry_data = {'date': '2016-05-01',
                                'time_worked': '00:30:00',
                                'description': 'test description'}

    def _verify_missing_task(self, response):
        self.assertEqual(response.status_code, 400)
        data = load_encoded_json(response.content)
        self.assertEqual(data['message'], 'No task for given id')
        self.assertEqual(data['error'], 400)

    def _verify_worker_not_assigned(self, response):
        self.assertEqual(response.status_code, 400)
        data = load_encoded_json(response.content)
        self.assertEqual(data['message'],
                         'Worker is not assigned to this task id.')
        self.assertEqual(data['error'], 400)

    def _verify_time_entries(self, data):
        for time_entry in data:
            serializer = TimeEntrySerializer(data=time_entry)
            self.assertTrue(serializer.is_valid())

    def test_time_entries_get(self):
        resp = self.request_client.get(self.url)
        data = load_encoded_json(resp.content)
        self.assertEqual(len(data), self.tasks.count())
        self._verify_time_entries(data)

    def test_time_entries_get_assignment_id(self):
        resp = self.request_client.get(self.url,
                                       data={'assignment': self.assignment.id})
        data = load_encoded_json(resp.content)
        self.assertEqual(len(data), 1)
        self._verify_time_entries(data)

    def test_time_entries_assignment_missing(self):
        resp = self.request_client.get(self.url, data={'assignment': '111'})
        data = load_encoded_json(resp.content)
        self.assertEqual(len(data), 0)

    def test_time_entries_post(self):
        resp = self.request_client.post(self.url,
                                        data=json.dumps(self.time_entry_data),
                                        content_type='application/json')
        data = load_encoded_json(resp.content)
        time_entry = TimeEntry.objects.get(id=data['id'])
        self.assertEqual(time_entry.worker, self.worker)
        self.assertIsNone(time_entry.assignment)
        self.assertIsNone(time_entry.timer_start_time)
        self.assertIsNone(time_entry.timer_stop_time)
        self.assertFalse(time_entry.is_deleted)
        self.assertEqual(time_entry.description, 'test description')
        self.assertEqual(time_entry.date, datetime.date(2016, 5, 1))
        self.assertEqual(time_entry.time_worked,
                         datetime.timedelta(minutes=30))

    def test_time_entries_post_invalid_data(self):
        self.time_entry_data['date'] = 'hi'  # Invalid date.
        resp = self.request_client.post(self.url,
                                        data=json.dumps(self.time_entry_data),
                                        content_type='application/json')
        self.assertEqual(resp.status_code, 400)

    def test_time_entries_post_cannot_set_worker(self):
        """
        Worker is a read-only field and writes to it will be ignored by the
        serializer.
        """
        worker = Worker.objects.exclude(id=self.worker.id).first()
        self.time_entry_data['worker'] = worker.id
        resp = self.request_client.post(self.url,
                                        data=json.dumps(self.time_entry_data),
                                        content_type='application/json')
        data = load_encoded_json(resp.content)
        time_entry = TimeEntry.objects.get(id=data['id'])
        self.assertNotEqual(time_entry.worker, worker)
        self.assertEqual(time_entry.worker, self.worker)

    def test_time_entry_delete(self):
        """
        Delete should mark instance as deleted, not delete instance.
        """
        time_entry = TimeEntry.objects.filter(worker=self.worker).first()
        self.request_client.delete(
            reverse('orchestra:orchestra:time_entry',
                    kwargs={'pk': time_entry.id}))
        self.assertFalse(TimeEntry.objects.filter(id=time_entry.id).exists())

    def test_time_entry_update(self):
        time_entry = TimeEntry.objects.filter(worker=self.worker).first()
        serializer = TimeEntrySerializer(time_entry)
        data = serializer.data
        data['time_worked'] = '10:00:00'  # Update time worked.
        self.request_client.put(
            reverse('orchestra:orchestra:time_entry',
                    kwargs={'pk': time_entry.id}),
            data=json.dumps(data),
            content_type='application/json')
        time_entry.refresh_from_db()
        self.assertEqual(time_entry.time_worked, datetime.timedelta(hours=10))

    def test_time_entry_update_invalid_data(self):
        time_entry = TimeEntry.objects.filter(worker=self.worker).first()
        serializer = TimeEntrySerializer(time_entry)
        data = serializer.data
        data['date'] = '2'  # Invalid date.
        resp = self.request_client.put(
            reverse('orchestra:orchestra:time_entry',
                    kwargs={'pk': time_entry.id}),
            data=json.dumps(data),
            content_type='application/json')
        self.assertEqual(resp.status_code, 400)


class TimerEndpointTests(EndpointTestCase):

    def setUp(self):
        super().setUp()
        user = UserFactory(username='test_user')
        self.worker = WorkerFactory(user=user)
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.time = timezone.now()

    @patch('orchestra.views.time_tracking.start_timer')
    def test_start_timer(self, mock_start):
        timer = TaskTimer(worker=self.worker, start_time=self.time)
        timer.save()
        mock_start.return_value = timer
        resp = self.request_client.post(
            reverse('orchestra:orchestra:start_timer'),
            data=json.dumps({'assignment': '11'}),
            content_type='application/json')
        mock_start.assert_called_with(self.worker, assignment_id='11')
        data = load_encoded_json(resp.content)
        serializer = TaskTimerSerializer(timer)
        self.assertEqual(data, serializer.data)

    @patch('orchestra.views.time_tracking.start_timer',
           side_effect=TaskAssignment.DoesNotExist)
    def test_start_timer_worker_not_assigned(self, mock_start):
        resp = self.request_client.post(
            reverse('orchestra:orchestra:start_timer'),
            data=json.dumps({'assignment': '11'}),
            content_type='application/json')
        self._verify_bad_request(resp,
                                 'Worker is not assigned to this task id.')

    @patch('orchestra.views.time_tracking.start_timer',
           side_effect=TimerError('test'))
    def test_start_timer_timer_error(self, mock_start):
        resp = self.request_client.post(
            reverse('orchestra:orchestra:start_timer'),
            data=json.dumps({'assignment': '11'}),
            content_type='application/json')
        self._verify_bad_request(resp, 'test')

    @patch('orchestra.views.time_tracking.stop_timer')
    def test_stop_timer(self, mock_stop):
        time_entry = TimeEntry(worker=self.worker, date=self.time.date(),
                               time_worked=datetime.timedelta(hours=1))
        time_entry.save()
        mock_stop.return_value = time_entry
        resp = self.request_client.post(
            reverse('orchestra:orchestra:stop_timer'),
            data=json.dumps({}),
            content_type='application/json')
        mock_stop.assert_called_with(self.worker)
        data = load_encoded_json(resp.content)
        serializer = TimeEntrySerializer(time_entry)
        self.assertEqual(data, serializer.data)

    @patch('orchestra.views.time_tracking.stop_timer',
           side_effect=TimerError('test'))
    def test_stop_timer_timer_error(self, mock_stop):
        resp = self.request_client.post(
            reverse('orchestra:orchestra:stop_timer'),
            data=json.dumps({}),
            content_type='application/json')
        self._verify_bad_request(resp, 'test')

    @patch('orchestra.views.time_tracking.get_timer_current_duration',
           return_value=datetime.timedelta(hours=1))
    def test_get_timer(self, mock_get):
        timer = TaskTimer(
            worker=self.worker,
            start_time=timezone.now())
        timer.save()
        resp = self.request_client.get(
            reverse('orchestra:orchestra:get_timer'))
        mock_get.assert_called_with(self.worker)
        data = load_encoded_json(resp.content)
        expected = TaskTimerSerializer(timer).data
        expected['time_worked'] = '1:00:00'
        self.assertEqual(data, expected)


class TestErrorViews(OrchestraTestCase):

    def assert_error_view(self, handler, status_code):
        request = MagicMock()
        response = handler(request)
        self.assertEqual(response.status_code, status_code)

    def test_bad_request(self):
        self.assert_error_view(bad_request, 400)

    def test_forbidden(self):
        self.assert_error_view(forbidden, 403)

    def test_not_found(self):
        self.assert_error_view(not_found, 404)

    def test_internal_server_error(self):
        self.assert_error_view(internal_server_error, 500)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0012_auto_20150701_1948'),
    ]

    operations = [
        migrations.RenameField(
            model_name='Project',
            old_name='process_slug',
            new_name='workflow_slug',
        ),
    ]



import django_filters

from orchestra.models import TimeEntry


class TimeEntryFilter(django_filters.FilterSet):
    min_date = django_filters.DateFilter(name='date', lookup_expr='gte')
    max_date = django_filters.DateFilter(name='date', lookup_expr='lte')

    class Meta:
        model = TimeEntry
        fields = ['min_date', 'max_date']



import kronos

from orchestra.communication.staffing import send_staffing_requests


@kronos.register('* * * * *')  # run every minute
def send_staffing_requests_periodically():
    send_staffing_requests()



from django.conf.urls import url

from orchestra.todos.views import TodoDetail
from orchestra.todos.views import TodoList
from orchestra.todos.views import TodoQADetail
from orchestra.todos.views import TodoQAList
from orchestra.todos.views import TodoListTemplateDetail
from orchestra.todos.views import TodoListTemplateList
import orchestra.todos.views as views

urlpatterns = [
    url(r'^todo/$',
        TodoList.as_view(), name='todos'),
    url(r'^todo/(?P<pk>[0-9]+)/$',
        TodoDetail.as_view(), name='todo'),
    url(r'^todo_qa/$',
        TodoQAList.as_view(), name='todo_qas'),
    url(r'^todo_qa/(?P<pk>[0-9]+)/$',
        TodoQADetail.as_view(), name='todo_qa'),
    url(r'^todolist_templates/$',
        TodoListTemplateList.as_view(), name='todolist_templates'),
    url(r'^todolist_template/(?P<pk>[0-9]+)/$',
        TodoListTemplateDetail.as_view(), name='todolist_template'),
    url(r'^update_todos_from_todolist_template/$',
        views.update_todos_from_todolist_template,
        name='update_todos_from_todolist_template'),
    url(r'^worker_task_recent_todo_qas/$',
        views.worker_task_recent_todo_qas,
        name='worker_task_recent_todo_qas'),
]



import math
from unittest.mock import patch

from django.test import TestCase

from orchestra.core.errors import S3UploadError
from orchestra.utils.s3 import MAX_UPLOAD_SIZE_MB
from orchestra.utils.s3 import upload_editor_image


@patch('orchestra.utils.s3.boto')
class BasicS3TestCase(TestCase):
    def test_upload_editor_image(self, boto_patch):
        # Sample unicode data string
        sample_data = 'data'.encode()

        # Attempt to upload incorrect image type
        with self.assertRaises(S3UploadError):
            upload_editor_image(sample_data, 'invalid_type')

        # Upload supported image type
        upload_editor_image(sample_data, 'image/jpeg')

        max_size_bytes = MAX_UPLOAD_SIZE_MB * 10 ** 6

        # Attempt to upload image larger than size limit
        sample_data = ('a' * math.floor(max_size_bytes * 1.0001)).encode()
        with self.assertRaises(S3UploadError):
            upload_editor_image(sample_data, 'image/jpeg')

        # Upload image smaller than size limit
        sample_data = ('a' * math.floor(max_size_bytes * 0.9999)).encode()
        upload_editor_image(sample_data, 'image/jpeg')



from orchestra.core.errors import TaskAssignmentError
from orchestra.models import Iteration
from orchestra.models import TaskAssignment


# TODO(kkamalov): move everything to model_mixins
def assignment_history(task):
    """
    Return all assignments for `task` ordered by `assignment_counter`.

    Args:
        task (orchestra.models.Task):
            The specified task object.

    Returns:
        assignment_history ([orchestra.models.TaskAssignment]):
            All assignments for `task` ordered by `assignment_counter`.
    """
    return task.assignments.order_by('assignment_counter')


def current_assignment(task):
    """
    Return the in-progress assignment for `task`.

    Args:
        task (orchestra.models.Task):
            The specified task object.

    Returns:
        current_assignment (orchestra.models.TaskAssignment):
            The in-progress assignment for `task`.
    """
    assignments = assignment_history(task)
    processing = task.assignments.filter(
        status=TaskAssignment.Status.PROCESSING)
    if not processing.exists():
        return assignments.last()
    else:
        if processing.count() > 1:
            raise TaskAssignmentError(
                'More than one processing assignment for task {}'.format(
                    task.id))
        else:
            return processing.first()


def get_latest_iteration(assignment):
    return assignment.iterations.order_by('start_datetime').last()


def get_iteration_history(task, reverse=False):
    order_expression = 'start_datetime'
    if reverse:
        order_expression = '-' + order_expression
    return (
        Iteration.objects.filter(assignment__task=task)
        .order_by(order_expression))


def all_workers(task):
    """
    Return all workers for a given task.

    Args:
        task (orchestra.models.Task):
            The specified task object.

    Returns:
        all_workers ([orchestra.models.Worker]):
            A list of all workers involved with `task`.
    """
    return [assignment.worker for assignment in assignment_history(task).all()]



from django.db import transaction
from django.utils import timezone

from orchestra.core.errors import TimerError
from orchestra.models import TaskAssignment
from orchestra.models import TaskTimer
from orchestra.models import TimeEntry


def get_timer_object(worker):
    """
    Returns TaskTimer object associated with worker, creates one if does not
    exist.
    """
    try:
        timer = worker.timer
    except TaskTimer.DoesNotExist:
        timer = TaskTimer.objects.create(worker=worker)
    return timer


def _reset_timer(timer):
    timer.assignment = None
    timer.start_time = None
    timer.stop_time = None
    timer.description = None
    timer.save()


def start_timer(worker, assignment_id=None):
    """
    Start timer for worker, and optionally task assignment id

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.
        assignment_id (int): optional
            The ID of the task assignment to be associated with the time entry.

    Returns:
        timer (orchestra.models.TaskTimer)
            newly created timer object

    Raises:
        orchestra.models.TaskAssignment.DoesNotExist:
            The specified worker is not assigned to the specified task.
        orchestra.core.errors.TimerError:
            Timer has already started.
    """
    timer = get_timer_object(worker)

    # Attach assignment if provided.
    if assignment_id:
        assignment = TaskAssignment.objects.get(id=assignment_id,
                                                worker=worker)
        timer.assignment = assignment

    if timer.start_time is not None:
        raise TimerError('Timer has already started')

    timer.start_time = timezone.now()
    timer.save()
    return timer


@transaction.atomic
def stop_timer(worker):
    """
    Stops timer for worker and creates a TimeEntry object.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.
        task_id (int):
            The ID of the task to be associated with the time entry.

    Returns:
        time_entry (orchestra.models.TimeEntry)
            newly created time entry from timer duration

    Raises:
        orchestra.core.errors.TimerError:
            Timer has not started.
    """
    timer = get_timer_object(worker)
    if timer.start_time is None:
        raise TimerError('Timer not started')
    timer.stop_time = timezone.now()
    timer.save()

    # Create TimeEntry object for timer.
    time_entry = TimeEntry.objects.create(
        worker=worker,
        date=timer.start_time.date(),
        time_worked=timer.stop_time - timer.start_time,
        assignment=timer.assignment,
        timer_start_time=timer.start_time,
        timer_stop_time=timer.stop_time,
        description=timer.description)

    # Reset timer
    _reset_timer(timer)

    return time_entry


def update_timer(worker, description, assignment_id):
    timer = get_timer_object(worker)
    timer.description = description
    if assignment_id is None:
        timer.assignment = None
    else:
        timer.assignment = TaskAssignment.objects.get(
            id=assignment_id, worker=worker)
    timer.save()


def get_timer_current_duration(worker):
    """
    Returns current time interval logged by timer for a worker.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.

    Returns:
        (datetime.timedelta)
            time interval since start of timer
    """
    timer = get_timer_object(worker)
    if timer.start_time is None:
        return None
    return timezone.now() - timer.start_time



import logging
import os
import re
import tempfile
from collections import Counter
from datetime import date

import requests
from django.conf import settings

from orchestra.google_apps.errors import FailedRequest
from orchestra.google_apps.errors import GoogleDriveError
from orchestra.google_apps.errors import InvalidUrlError
from orchestra.google_apps.permissions import read_with_link_permission
from orchestra.google_apps.permissions import write_with_link_permission
from orchestra.google_apps.service import Service
from orchestra.utils.common_regex import image_file_regex
from orchestra.utils.decorators import run_if

logger = logging.getLogger(__name__)
_image_mimetype_regex = re.compile('(image/(?:jpg|jpeg|gif|png|svg))',
                                   re.IGNORECASE)
TEAM_MESSAGES_TEMPLATE_ID = '1d0kIgq8G_Su6j5abP-tP6yJ2sp-sFDk6vZiREil0_70'


def _get_image_mimetype(response, title):
    """
    Provided http response and an image title
    generate an image mimetype.
    """
    if (response.headers.get('content-type') and
        _image_mimetype_regex.search(response.headers
                                     .get('content-type'))):
        return response.headers.get('content-type')
    extension = title.split('.')[-1]
    return 'image/{}'.format(extension)


@run_if('GOOGLE_APPS')
def add_image(service, folder_id, url):
    """
    Add image to a folder.

    Args:
      service: Drive API service instance.
      folder_id: ID of a folder where the image will be stored
      url: url to the original image

    Returns:
      Metadata to an uploaded file
    """

    response = requests.get(url, stream=True)
    if response.status_code != 200:
        raise FailedRequest('Unable to successfully retrieve '
                            'an image: %s', (url))

    temp = tempfile.NamedTemporaryFile(mode='wb', delete=False)
    for chunk in response.iter_content(1024):
        temp.write(chunk)

    title_regex = image_file_regex.search(response.url)
    if title_regex is None:
        raise InvalidUrlError('Url is not for an image.')

    title = title_regex.group()
    mimetype = _get_image_mimetype(response, title)
    temp.close()
    google_image = service.insert_file(title,
                                       'image',
                                       folder_id,
                                       mimetype,
                                       temp.name)
    os.unlink(temp.name)
    return google_image


@run_if('GOOGLE_APPS')
def create_media_folder_with_images(parent_id, image_links, folder_name):
    """
    Given a folder name and a list of image links create a new
    google drive folder with images in it.
    """
    service = Service(settings.GOOGLE_P12_PATH,
                      settings.GOOGLE_SERVICE_EMAIL)
    folder = create_folder_with_permissions(parent_id,
                                            folder_name,
                                            [read_with_link_permission])
    folder_id = folder['id']
    counter = Counter()
    for image_link in image_links:
        try:
            image = add_image(service, folder_id, image_link)
            counter['uploaded_images'] += 1
            logger.info('Image has been uploaded %s', image)
        except (InvalidUrlError, FailedRequest):
            counter['not_uploaded_images'] += 1
            logger.exception('Failed to retrieve image from %s',
                             image_link)
    return {'folder': folder,
            'image_counter': counter}


@run_if('GOOGLE_APPS')
def create_folder_with_permissions(parent_id, folder_name, permissions=None):
    """
    Create drive folder in the specified location with given permissions.
    """
    service = Service(settings.GOOGLE_P12_PATH,
                      settings.GOOGLE_SERVICE_EMAIL)
    folder = service.insert_folder(folder_name, parent_id)
    if folder is None:
        raise GoogleDriveError('Could not create a folder')

    permissions = permissions or []
    for permission in permissions:
        service.add_permission(folder.get('id'),
                               permission)

    return folder


@run_if('GOOGLE_APPS')
def create_project_google_folder(project):
    """
    Create drive folder for project information
    """
    today = date.today().strftime('%Y-%m-%d')
    parent_id = (project.project_data.get('client_folder_id') or
                 settings.GOOGLE_PROJECT_ROOT_ID)
    folder = create_folder_with_permissions(
        parent_id,
        ' '.join((today, project.short_description)),
        [write_with_link_permission])
    folder_id = folder.get('id')
    project.project_data['project_folder_id'] = folder_id
    project.team_messages_url = create_document_from_template(
        TEAM_MESSAGES_TEMPLATE_ID,
        'Team Messages',
        [folder_id],
        [write_with_link_permission])['alternateLink']
    project.save()
    return folder


@run_if('GOOGLE_APPS')
def create_document_from_template(template_id, name,
                                  parent_ids=None, permissions=None):
    service = Service(settings.GOOGLE_P12_PATH,
                      settings.GOOGLE_SERVICE_EMAIL)
    upload_info = service.copy_file(template_id, name,
                                    parent_ids=parent_ids)
    if upload_info is None:
        raise GoogleDriveError('Could not create document {}'.format(name))

    logger.info(upload_info)
    document_id = upload_info.get('id')

    permissions = permissions or []
    for permission in permissions:
        service.add_permission(document_id, permission)

    upload_info['status'] = 'success'
    upload_info['id'] = document_id
    return upload_info


@run_if('GOOGLE_APPS')
def download_file(file_metadata):
    """Download a file from a google drive folder.

    Args:
        file_metadata (dict):
            A Google Apps API file resource.

    Returns:
        file_contents (str):
            A buffer containing the raw binary contents of the file.
        title (str):
            The title of the file.
        mimetype(str):
            The mimetype of the file.
    """

    service = Service(settings.GOOGLE_P12_PATH,
                      settings.GOOGLE_SERVICE_EMAIL)
    mimetype = file_metadata['mimeType']
    title = file_metadata['title']
    return service.get_file_content(file_metadata['id']), title, mimetype


@run_if('GOOGLE_APPS')
def upload_file(parent_id, file_path, title, description, mimetype):
    """Upload a file to a google drive folder.

    Args:
        parent_id (str):
            Identifier for the drive folder to upload to.
        file_path (str):
            Local file path to the file to upload.
        title (str):
            Title for the uploaded document.
        description (str):
            A description of the file to upload.
        mimetype (str):
            Mimetype of the uploaded content.

    Returns:
        file_metadata (dict):
            A Google Apps File resource with metadata about the uploaded file.
    """

    service = Service(settings.GOOGLE_P12_PATH,
                      settings.GOOGLE_SERVICE_EMAIL)
    file_metadata = service.insert_file(
        title,
        description,
        parent_id,
        mimetype,
        file_path
    )
    return file_metadata



import jsl


class PredicateSchema(jsl.Document):
    """
    A predicate schema

    Attributes:
        operator (str):
            Specify the compare operator of predicate.
            Supported predicates are >, <, >=, <=, !=, and ==.
        value (number, bool, sring, or None):
            The value to compare with of the predicate.
    """
    operator = jsl.StringField(required=True, pattern='[!=><]=|[><]')
    value = jsl.AnyOfField([
        jsl.NumberField(),
        jsl.BooleanField(),
        jsl.StringField(),
        jsl.NullField()], required=True)


class TodoSchema(jsl.Document):
    """
    A Todo schema

    Attributes:
        id (int):
            A unique id for the todo.
        description (str):
            A text description of the todo.
        items (array):
            An array of sub-todos of this todo.
        skip_if (array):
            An array of conditions to skip this todo. If any of the
            condition is true, the todo is skipped. Each condition is a
            dictionary of attributes and predicates which get ANDed together.
        remove_if (array):
            An array of conditions to remove this todo. If any of the
            condition is true, the todo is removed. Each condition is a
            dictionary of attributes and predicates which get ANDed together.
    """
    id = jsl.IntField(required=True)
    description = jsl.StringField(required=True)
    items = jsl.ArrayField(jsl.DocumentField('TodoSchema'))
    skip_if = jsl.ArrayField(
        jsl.DictField(
            pattern_properties={'.*': jsl.DocumentField('PredicateSchema')}))
    remove_if = jsl.ArrayField(
        jsl.DictField(
            pattern_properties={'.*': jsl.DocumentField('PredicateSchema')}))


class TodoListSchema(jsl.Document):
    items = jsl.ArrayField(jsl.DocumentField('TodoSchema'))


class TodoActionSchema(jsl.Document):
    """
    A todo action schema

    Attributes:
        action (str):
            Specify the action taken by the worker on the given todo.
            Examples: skip, unskip, complete, incomplete
        datetime (datetime):
            The time the action was taken.
        step_slug (str):
            Unique identifier for the workflow step that the task represents.
    """
    action = jsl.StringField(required=True)
    datetime = jsl.DateTimeField(required=True)
    step_slug = jsl.StringField(required=True)


class TodoActionListSchema(jsl.Document):
    actions = jsl.ArrayField(jsl.DocumentField('TodoActionSchema'))



from django.db import transaction
from django.utils import timezone
import logging
import operator
from pydoc import locate

from orchestra.models import TodoListTemplate
from orchestra.models import Todo
from orchestra.models import Task

logger = logging.getLogger(__name__)

OPERATORS = {
    '<': operator.lt,
    '<=': operator.le,
    '==': operator.eq,
    '!=': operator.ne,
    '>=': operator.ge,
    '>': operator.gt
}


@transaction.atomic
def add_todolist_template(todolist_template_slug, task_id):
    todolist_template = TodoListTemplate.objects.get(
        slug=todolist_template_slug)

    task = Task.objects.get(id=task_id)
    template_todos = todolist_template.todos.get('items', [])
    root_todo = Todo(
        task=task,
        description=todolist_template.name,
        template=todolist_template
    )
    root_todo.save()

    cond_props = {}
    path = todolist_template.conditional_property_function.get(
        'path', None)
    if path:
        try:
            get_cond_props = locate(path)
            cond_props = get_cond_props(task.project)
        except Exception:
            logger.exception('Invalid conditional function path.')
    for template_todo in template_todos:
        _add_template_todo(
            template_todo, todolist_template, root_todo, task, cond_props)


def _to_exclude(props, conditions):
    """
    The conditions is it a list of conditions that get ORed together,
    with predicates in each dictionary getting ANDed.
    """
    any_condition_true = False

    for condition in conditions:
        all_props_true = len(condition) > 0
        for prop, predicate in condition.items():
            current_value = props.get(prop)
            compared_to_value = predicate['value']
            compare = OPERATORS[predicate['operator']]
            all_props_true = (
                all_props_true and
                compare(current_value, compared_to_value))
        any_condition_true = any_condition_true or all_props_true

    return any_condition_true


def _add_template_todo(
        template_todo, todolist_template,
        parent_todo, task, conditional_props):
    remove = _to_exclude(conditional_props, template_todo.get('remove_if', []))
    if not remove:
        if parent_todo.skipped_datetime:
            skipped_datetime = parent_todo.skipped_datetime
        else:
            to_skip = _to_exclude(
                    conditional_props, template_todo.get('skip_if', []))
            skipped_datetime = timezone.now() if to_skip else None

        todo = Todo(
            task=task,
            description=template_todo['description'],
            template=todolist_template,
            parent_todo=parent_todo,
            skipped_datetime=skipped_datetime
        )
        todo.save()
        for template_todo_item in template_todo.get('items', []):
            _add_template_todo(
                template_todo_item, todolist_template, todo,
                task, conditional_props)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.9 on 2016-08-06 19:54
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0066_remove_taskassignment_integrity_constraint'),
    ]

    operations = [
        migrations.AddField(
            model_name='staffbotrequest',
            name='last_inquiry_sent',
            field=models.DateTimeField(null=True),
        ),
    ]



from pydoc import locate

from django.db.models import Q

from orchestra.core.errors import ModelSaveError
from orchestra.json_schemas.validation import validate_json
from orchestra.json_schemas.tags import TagListSchema


class WorkflowMixin(object):

    def __str__(self):
        return self.slug


class WorkflowVersionMixin(object):

    def __str__(self):
        return '{} - {}'.format(self.workflow.slug, self.slug)


class CertificationMixin(object):

    def __str__(self):
        return '{} - {}'.format(self.slug, self.workflow.slug)


class SanityCheckMixin(object):

    def __str__(self):
        return '{} - {} (created: {}, handled: {})'.format(
            self.project,
            self.check_slug,
            self.created_at,
            self.handled_at)


class StepMixin(object):

    def __str__(self):
        return '{} - {} - {}'.format(self.slug, self.workflow_version.slug,
                                     self.workflow_version.workflow.slug)


class WorkerMixin(object):

    def has_certificate(self, certification, role):
        from orchestra.models import WorkerCertification
        return WorkerCertification.objects.filter(
            worker=self,
            certification=certification,
            role=role
        ).exists()

    def is_reviewer(self, certification):
        from orchestra.models import WorkerCertification
        return self.has_certificate(
            certification,
            WorkerCertification.Role.REVIEWER
        )

    def is_entry_level(self, certification):
        from orchestra.models import WorkerCertification
        return self.has_certificate(
            certification,
            WorkerCertification.Role.ENTRY_LEVEL
        )

    def is_project_admin(self):
        from orchestra.interface_api.project_management.decorators import (
            is_project_admin)
        return is_project_admin(self.user)

    def __str__(self):
        return '{} - @{} -{}'.format(
            self.user.username,
            self.slack_username,
            self.phone
        )

    def formatted_slack_username(self):
        return '<@{}>'.format(self.slack_user_id)


class WorkerCertificationMixin(object):

    def __str__(self):
        return '{} - {} - {} - {} - {}'.format(
            self.worker.user.username, self.certification.slug,
            self.certification.workflow.slug,
            dict(self.TASK_CLASS_CHOICES)[self.task_class],
            dict(self.ROLE_CHOICES)[self.role])

    def save(self, *args, **kwargs):
        if self.role == self.Role.REVIEWER:
            if not (type(self).objects
                    .filter(worker=self.worker, task_class=self.task_class,
                            certification=self.certification,
                            role=self.Role.ENTRY_LEVEL)
                    .exists()):
                raise ModelSaveError('You are trying to add a reviewer '
                                     'certification ({}) for a worker without '
                                     'an entry-level certification'
                                     .format(self))
        super().save(*args, **kwargs)


class ProjectMixin(object):

    def __str__(self):
        return '{} ({})'.format(str(self.workflow_version.slug),
                                self.short_description)


class TaskMixin(object):

    def is_worker_assigned(self, worker):
        """
        Check if specified worker is assigned to the given task.

        Args:
            worker (orchestra.models.Worker):
                The specified worker object.

        Returns:
            worker_assigned_to_task (bool):
                True if worker has existing assignment for the given task.
        """
        return self.assignments.filter(worker=worker).exists()

    def get_detailed_description(self, extra_kwargs=None):
        """
        This function uses a step's `description_function` field to generate
        dynamic text describing a step within a task.
        Args:
            extra_kwargs (dict):
                Additional (dynamic) kwargs that will be passed to the
                description function.
        Returns:
            detailed_description (str):
                Dynamic message describing the task.
        """
        from orchestra.utils.task_lifecycle import (
            get_task_details)
        path = self.step.detailed_description_function.get('path')
        kwargs = self.step.detailed_description_function.get('kwargs', {})
        if extra_kwargs is not None:
            kwargs.update(extra_kwargs)
        try:
            function = locate(path)
            return function(get_task_details(self.id), **kwargs)
        except Exception:
            return ''

    def save(self, *args, **kwargs):
        validate_json('tags', TagListSchema,
                      getattr(self, 'tags', None))
        super().save(*args, **kwargs)

    def __str__(self):
        return '{} - {}'.format(str(self.project), str(self.step.slug))


class TaskAssignmentMixin(object):

    def save(self, *args, **kwargs):
        if self.task.step.is_human:
            if self.worker is None:
                raise ModelSaveError('Worker has to be present '
                                     'if worker type is Human')
        else:
            if self.worker is not None:
                raise ModelSaveError('Worker should not be assigned '
                                     'if worker type is Machine')

        super().save(*args, **kwargs)

    def is_entry_level(self):
        return self.assignment_counter == 0

    def is_reviewer(self):
        return self.assignment_counter > 0

    def __str__(self):
        return '{} - {} - {}'.format(
            str(self.task), self.assignment_counter, str(self.worker))


class TodoMixin(object):

    def __str__(self):
        return '{} - {} ({})'.format(
            self.task,
            self.description,
            self.completed)


class TodoQAMixin(object):

    def __str__(self):
        return '{} - {} ({})'.format(
            self.todo,
            self.comment,
            self.approved)


class TodoListTemplateMixin(object):

    def __str__(self):
        return '{} - {}'.format(
            self.name, self.description)


class PayRateMixin(object):

    def __str__(self):
        return '{} ({} - {})'.format(
            self.worker, self.start_date, self.end_date or 'now')

    def save(self, *args, **kwargs):
        if self.end_date and self.end_date < self.start_date:
            raise ModelSaveError('end_date must be greater than '
                                 'start_date')

        if self.end_date is None:
            # If end_date is None, need to check that no other PayRates have
            # end_date is None, nor do they overlap.
            if type(self).objects.exclude(id=self.id).filter(
                    (Q(end_date__gte=self.start_date) |
                     Q(end_date__isnull=True)),
                    worker=self.worker).exists():
                raise ModelSaveError(
                    'Date range overlaps with existing PayRate entry')
        else:
            # If end_date is not None, need to check if other PayRates overlap.
            if (type(self).objects.exclude(id=self.id).filter(
                    start_date__lte=self.end_date,
                    end_date__isnull=True,
                    worker=self.worker).exists() or
                type(self).objects.exclude(id=self.id).filter(
                    (Q(start_date__lte=self.end_date) &
                     Q(end_date__gte=self.start_date)),
                    worker=self.worker).exists()):
                raise ModelSaveError(
                    'Date range overlaps with existing PayRate entry')
        super().save(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0002_auto_20141229_1543'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='short_description',
            field=models.CharField(default='', max_length=100),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='project',
            name='task_class',
            field=models.IntegerField(
                choices=[(0, 'Training tasks'), (1, 'A real task')]),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='step',
            name='worker_type',
            field=models.IntegerField(choices=[(0, 'Human'), (1, 'Machine')]),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='task',
            name='status',
            field=models.IntegerField(choices=[(0, 'Awaiting Processing'), (1, 'Processing'), (
                2, 'Pending Review'), (3, 'Reviewing'), (4, 'Post-review Processing'), (5, 'Complete')]),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='status',
            field=models.IntegerField(
                choices=[(0, 'Procesing'), (1, 'Submitted')]),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='workercertification',
            name='role',
            field=models.IntegerField(
                choices=[(0, 'Entry-level'), (1, 'Reviewer')]),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='workercertification',
            name='task_class',
            field=models.IntegerField(
                choices=[(0, 'Training tasks'), (1, 'A real task')]),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0003_auto_20141229_1610'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskassignment',
            name='task',
            field=models.ForeignKey(
                on_delete=models.CASCADE, default=0, to='orchestra.Task'),
            preserve_default=False,
        ),
    ]



import logging
import os
import tempfile

from django.conf import settings
from PIL import Image

from orchestra.google_apps.convenience import create_folder_with_permissions
from orchestra.google_apps.convenience import download_file
from orchestra.google_apps.convenience import upload_file
from orchestra.google_apps.permissions import write_with_link_permission
from orchestra.google_apps.service import Service

logger = logging.getLogger(__name__)


def autoadjust_photos(project_data, prerequisites):
    """Resize all images in a google drive directory."""
    task_data = {}
    parent_folder_id = project_data['project_folder_id']

    # Create a directory to output the photos
    output_folder = create_folder_with_permissions(
        parent_folder_id,
        'Processed Photos',
        permissions=[write_with_link_permission],
    )
    task_data['processed_photo_folder'] = output_folder['id']

    # List the existing photos
    raw_photo_folder_id = (prerequisites
                           .get('document_creation')
                           .get('raw_photo_folder'))
    service = Service(settings.GOOGLE_P12_PATH,
                      settings.GOOGLE_SERVICE_EMAIL)
    photos_metadata = service.list_folder(raw_photo_folder_id)

    # Iterate over the input photos and process them.
    task_data['photos_for_caption'] = []
    for photo_metadata in photos_metadata:
        photo, title, mimetype = download_file(photo_metadata)
        adjusted_photo_tmpfile = adjust_photo(photo)
        upload = upload_file(
            task_data['processed_photo_folder'],
            adjusted_photo_tmpfile.name,
            title,
            'image',
            mimetype
        )
        os.unlink(adjusted_photo_tmpfile.name)
        embed_link = upload['webContentLink'].replace('&export=download', '')
        task_data['photos_for_caption'].append(embed_link)
    return task_data


def adjust_photo(photo):

    # Write the photo to a temporary file
    temp = tempfile.NamedTemporaryFile(mode='wb', delete=False)
    temp.write(photo)

    # Open it up and play with it in PIL
    im = Image.open(temp.name)
    im = im.convert('L')  # convert to greyscale

    # Save it back out into a new temporary file
    os.unlink(temp.name)
    temp2 = tempfile.NamedTemporaryFile(mode='wb', delete=False)
    im.save(temp2, format='jpeg')
    return temp2



from unittest.mock import MagicMock

import slacker
from django.conf import settings

# Groups that are not dynamically created during testing
PREEXISTING_GROUPS = [settings.SLACK_INTERNAL_NOTIFICATION_CHANNEL]


MOCK_SLACK_API_DATA = {
    'groups': {},
    'users': {},
    'invited': [],
}


class MockSlacker(MagicMock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = MOCK_SLACK_API_DATA
        self.groups = Groups()
        self.chat = Chat()
        self.users = Users()
        self.populate_preexisting_groups()

    def populate_preexisting_groups(self):
        for group_name in PREEXISTING_GROUPS:
            self.groups.create(group_name)

    def get_messages(self, group_id):
        return MOCK_SLACK_API_DATA['groups'][group_id]['messages']

    def clear(self):
        MOCK_SLACK_API_DATA['groups'].clear()
        MOCK_SLACK_API_DATA['users'].clear()
        MOCK_SLACK_API_DATA['invited'][:] = []
        self.populate_preexisting_groups()


class BaseAPI(object):

    class Response(object):
        def __init__(self, body):
            self.body = body

    def _group_exists(self, group_id):
        return MOCK_SLACK_API_DATA['groups'].get(group_id, None) is not None

    def _user_exists(self, user_id):
        return MOCK_SLACK_API_DATA['users'].get(user_id, None) is not None

    def _validate_group(self, group_id):
        if not group_id or not self._group_exists(group_id):
            raise slacker.Error('Group not found.')

    def _validate_user(self, user_id):
        if not user_id or not self._user_exists(user_id):
            raise slacker.Error('User not found.')

    def post(self, command, data):
        if command == 'users.admin.invite':
            user_email = data.get('email', None)
            if user_email:
                if user_email not in MOCK_SLACK_API_DATA['invited']:
                    MOCK_SLACK_API_DATA['invited'].append(user_email)
            else:
                raise slacker.Error('User email was not provided')

    def accept_invite(self, email, username):
        """
        Imitate the user signing up for a slack account after invitation.
        """
        MOCK_SLACK_API_DATA['invited'].remove(email)
        # Mock user id is the same as username
        user_id = username
        MOCK_SLACK_API_DATA['users'][user_id] = {
            'email': email,
            'name': username
        }


class Groups(BaseAPI):
    def create(self, group_name):
        group_id = str(len(MOCK_SLACK_API_DATA['groups']))
        MOCK_SLACK_API_DATA['groups'][group_id] = {
            'id': group_id,
            'users': [],
            'messages': [],
            'topic': None,
            'purpose': None,
            'name': group_name.strip('#'),
        }
        return self.Response({
            'group': MOCK_SLACK_API_DATA['groups'][group_id]
        })

    def invite(self, group_id, user_id):
        self._validate_group(group_id=group_id)
        self._validate_user(user_id=user_id)

        already_in_group = True
        if user_id not in MOCK_SLACK_API_DATA['groups'][group_id]['users']:
            # Slacker API does not raise an error if user already present
            MOCK_SLACK_API_DATA['groups'][group_id]['users'].append(user_id)
            already_in_group = False
        return self.Response({'already_in_group': already_in_group})

    def info(self, group_id):
        self._validate_group(group_id=group_id)
        return self.Response({
            'ok': True,
            'group': {
                'id': group_id,
                'is_archived': True
            }
        })

    def kick(self, group_id, user_id):
        self._validate_group(group_id=group_id)
        self._validate_user(user_id=user_id)

        if user_id not in MOCK_SLACK_API_DATA['groups'][group_id]['users']:
            raise slacker.Error('User does not belong to group.')

        # Slacker API does not raise an error if user already present
        MOCK_SLACK_API_DATA['groups'][group_id]['users'].remove(user_id)

    def set_topic(self, group_id, topic):
        self._validate_group(group_id=group_id)
        MOCK_SLACK_API_DATA['groups'][group_id]['topic'] = topic

    def set_purpose(self, group_id, purpose):
        if not self._group_exists(group_id):
            raise slacker.Error('Group not found.')

        MOCK_SLACK_API_DATA['groups'][group_id]['purpose'] = purpose

    def list(self):
        return self.Response({
            'ok': True,
            'groups': list(MOCK_SLACK_API_DATA['groups'].values())
        })

    def archive(self, group_id):
        self._validate_group(group_id=group_id)
        return self.Response({
            'ok': True
        })

    def unarchive(self, group_id):
        self._validate_group(group_id=group_id)
        return self.Response({
            'ok': True
        })


class Chat(BaseAPI):
    def post_message(self, group_identifier, text, parse='none'):
        if group_identifier.startswith('#'):
            groups = [
                group for group in MOCK_SLACK_API_DATA['groups'].values()
                if group['name'] == group_identifier.strip('#')]
            group_identifier = groups[0]['id']
        self._validate_group(group_id=group_identifier)
        MOCK_SLACK_API_DATA[
            'groups'][group_identifier]['messages'].append(text)


class Users(BaseAPI):
    def get_user_id(self, user_name):
        for uid, uname in MOCK_SLACK_API_DATA['users'].items():
            if uname == user_name:
                return uid
        return None
